#!/usr/bin/env node
import { command, cli } from 'cleye';
import _ from 'lodash';
import p, { select, text, confirm, outro, cancel, intro, group } from '@clack/prompts';
import { dim, red, green, blue, lightGreen, bgYellow, black } from 'kolorist';
import dedent from 'dedent';
import { promisify } from 'node:util';
import { exec, spawn } from 'node:child_process';
import fs from 'node:fs/promises';
import os from 'os';
import ini from 'ini';
import path from 'path';

var name = "commitizen-cli";
var version = "2.0.0";
var author = {
	name: "iamando",
	email: "iamando@tuta.io"
};

const PACKAGE_NAME = name;
const VERSION = version;
const AUTHOR = author;
const CANCELED_OP_MSG = "Operation Canceled";

const type = () => select({
  message: "Choose commit type",
  options: [
    { label: "Feature", value: "feature" },
    { label: "Bugfix", value: "bugfix" },
    { label: "Hotfix", value: "hotfix" },
    { label: "Chore", value: "chore" },
    { label: "Epic", value: "epic" },
    { label: "Design", value: "design" },
    { label: "Experiment", value: "experiment" },
    { label: "Documentation", value: "documentation" }
  ]
});
const commit = () => text({
  message: "Insert commit message",
  validate: (value) => {
    if (_.isEmpty(value))
      return "Commit message required";
  }
});
const ticket = async () => {
  const hasTicket = await confirm({
    message: "Has ticket ?",
    initialValue: false
  });
  if (!hasTicket)
    return;
  await text({
    message: "Insert ticket name",
    validate: (value) => {
      if (_.isEmpty(value))
        return "Ticket name required";
    }
  });
};
const emoji = () => confirm({
  message: "Use emoji on commit type ?",
  initialValue: false
});

const log = ({
  type,
  msg,
  isConsole = true,
  newLine = true
}) => {
  switch (type) {
    case "info":
      return isConsole ? console.info(`${newLine ? "\n" : ""} ${blue("\u2754")} ${msg}`) : `${newLine ? "\n" : ""} ${blue("\u2754")} ${msg}`;
    case "success":
      return isConsole ? console.log(`${newLine ? "\n" : ""} ${green("\u2714")} ${msg}`) : `${newLine ? "\n" : ""} ${green("\u2714")} ${msg}`;
    case "error":
      return isConsole ? console.error(` ${newLine ? "\n" : ""} ${red("\u274C")} ${msg}`) : `${newLine ? "\n" : ""} ${red("\u274C")} ${msg}`;
    default:
      return isConsole ? console.log(`${newLine ? "\n" : ""} ${dim("\u2755")} ${msg}`) : `${newLine ? "\n" : ""} ${dim("\u2755")} ${msg}`;
  }
};

class CliError extends Error {
}
const indent = " ".repeat(4);
const handleCliError = (error) => {
  if (error instanceof Error && !(error instanceof CliError)) {
    if (error.stack) {
      log({
        type: "error",
        msg: error.stack.split("\n").slice(1).join("\n")
      });
    }
    log({
      type: "error",
      msg: `
${indent}${dim(`${PACKAGE_NAME} v${VERSION}`)}`
    });
    log({
      type: "error",
      msg: `
${indent}Please open a Bug report with the information above:`
    });
    log({
      type: "error",
      msg: `${indent}https://github.com/${AUTHOR.name}/${PACKAGE_NAME}/issues/new`
    });
  }
};

const execa = promisify(exec);
const gitStatus = async () => {
  const { stdout: stdoutStatus, stderr: stderrStatus } = await execa(
    "git status"
  );
  return { stdoutStatus, stderrStatus };
};
const gitAdd = async () => {
  const { stdout: stdoutAdd, stderr: stderrAdd } = await execa("git add .");
  return { stdoutAdd, stderrAdd };
};
const gitCommit = async ({ commit }) => {
  const { stdout: stdoutCommit, stderr: stderrCommit } = await execa(
    `git commit -m "${commit}"`
  );
  return { stdoutCommit, stderrCommit };
};
const isTreeClean = async () => {
  const { stdoutStatus, stderrStatus } = await gitStatus();
  if (stderrStatus)
    throw new CliError(`An error occured: ${stderrStatus}`);
  if (stdoutStatus.includes("nothing to commit, working tree clean")) {
    outro(lightGreen("Nothing to commit, working tree clean \u{1F9F9}"));
    process.exit(1);
  }
};

const { hasOwnProperty } = Object.prototype;
const hasOwn = (object, key) => hasOwnProperty.call(object, key);
const configParsers = {
  PATTERN(pattern) {
    if (!pattern)
      return "";
    return pattern;
  }
};
const configPath = path.join(os.homedir(), ".commitizen-cli");
const fileExists = (filePath) => fs.lstat(filePath).then(
  () => true,
  () => false
);
const readConfigFile = async () => {
  const configExists = await fileExists(configPath);
  if (!configExists) {
    return /* @__PURE__ */ Object.create(null);
  }
  const configString = await fs.readFile(configPath, "utf8");
  return ini.parse(configString);
};
const getConfig = async (cliConfig) => {
  const config = await readConfigFile();
  const parsedConfig = {};
  for (const key of Object.keys(configParsers)) {
    const parser = configParsers[key];
    const value = cliConfig?.[key] ?? config[key];
    parsedConfig[key] = parser(value);
  }
  return parsedConfig;
};
const setConfigs = async (keyValues) => {
  const config = await readConfigFile();
  for (const [key, value] of keyValues) {
    if (!hasOwn(configParsers, key)) {
      throw new CliError(`Invalid config property: ${key}`);
    }
    const parsed = configParsers[key](value);
    config[key] = parsed;
  }
  await fs.writeFile(configPath, ini.stringify(config), "utf8");
};

const emojiByType = {
  feature: "\u2728",
  bug: "\u{1F41B}",
  hotfix: "\u{1F691}",
  chore: "\u{1F6E0}\uFE0F",
  epic: "\u{1F4CC}",
  design: "\u{1F3A8}",
  experiment: "\u{1F9EA}",
  documentation: "\u{1F4DD}"
};
const pattern = async ({
  type,
  commit,
  ticket,
  emoji
}) => {
  let defaultPattern;
  const { PATTERN: configPattern } = await getConfig();
  if (_.isEmpty(configPattern)) {
    ticket ? defaultPattern = ":emoji :type(:ticket): :commit" : defaultPattern = ":emoji :type: :commit";
  } else {
    defaultPattern = configPattern;
  }
  const hasPattern = await confirm({
    message: "Has specific pattern ?",
    initialValue: false
  });
  if (p.isCancel(hasPattern)) {
    cancel(CANCELED_OP_MSG);
    process.exit(0);
  }
  if (!hasPattern)
    return defaultPattern;
  const pattern2 = await text({
    message: "Insert specific pattern",
    placeholder: "Example: :type(:ticket): :commit",
    defaultValue: defaultPattern,
    validate: (value) => {
      if (_.isEmpty(value))
        return "Specific pattern required";
      if (!hasPatternKeys({ type, commit, ticket, emoji, pattern: value }))
        return "Pattern key not invalid (ex: :type, :ticket, :commit)";
    }
  });
  if (p.isCancel(pattern2)) {
    cancel(CANCELED_OP_MSG);
    process.exit(0);
  }
  const savePatterConfig = await confirm({
    message: "Save pattern to config ?",
    initialValue: true
  });
  if (p.isCancel(savePatterConfig)) {
    cancel(CANCELED_OP_MSG);
    process.exit(0);
  }
  setConfigs([["PATTERN", pattern2]]);
};
const hasPatternKeys = ({
  type,
  commit,
  ticket,
  emoji,
  pattern: pattern2
}) => {
  const patternKeys = [];
  if (!_.isEmpty(type))
    patternKeys.push(":type");
  if (!_.isEmpty(commit))
    patternKeys.push(":commit");
  if (!_.isEmpty(ticket))
    patternKeys.push(":ticket");
  if (emoji)
    patternKeys.push(":emoji");
  const containKeys = patternKeys.every((key) => pattern2.includes(key));
  return containKeys;
};
const replace = ({
  pattern: pattern2,
  values
}) => {
  for (const value in values) {
    pattern2 = pattern2?.replace(
      new RegExp(value, "g"),
      values[value]
    );
  }
  return pattern2;
};
const transform = async ({
  type,
  commit,
  ticket,
  emoji,
  pattern: pattern2
}) => {
  const patternKeys = {
    ":type": type,
    ":ticket": ticket ?? "",
    ":commit": commit,
    ":emoji": emoji ? emojiByType[type] : ""
  };
  const message = replace({ pattern: pattern2, values: patternKeys });
  return message;
};

const commiter = async () => {
  intro(bgYellow(black("Commitizen CLI")));
  await isTreeClean();
  const values = await group(
    {
      type: () => type(),
      commit: () => commit(),
      ticket: () => ticket(),
      emoji: () => emoji()
    },
    {
      onCancel: () => {
        cancel(CANCELED_OP_MSG);
        process.exit(0);
      }
    }
  );
  try {
    const blueprint = await pattern({
      type: values.type,
      commit: values.commit,
      ticket: values.ticket,
      emoji: values.emoji
    });
    if (_.isEmpty(blueprint))
      return;
    const commit2 = await transform({
      type: values.type,
      commit: values.commit,
      ticket: values.ticket,
      emoji: values.emoji,
      pattern: blueprint
    });
    const { stdoutStatus, stderrStatus } = await gitStatus();
    if (stderrStatus)
      throw new CliError(`An error occured: ${stderrStatus}`);
    if (stdoutStatus.includes("no changes added to commit") || stdoutStatus.includes(
      "nothing added to commit but untracked files present"
    )) {
      const type2 = stdoutStatus.includes("no changes added to commit") ? "modified" : "untracked";
      const addStagedFiles = await confirm({
        message: `No changes added to commit, would you like to add ${type2} files ?`,
        initialValue: true
      });
      if (addStagedFiles) {
        const { stderrAdd } = await gitAdd();
        if (stderrAdd)
          throw new CliError(`An error occured: ${stderrAdd}`);
        const { stderrCommit: stderrCommit2 } = await gitCommit({ commit: commit2.trim() });
        if (stderrCommit2)
          throw new CliError(`An error occured: ${stderrCommit2}`);
        outro(dedent`
        You're all set 🎉

        use "git push" to publish your local commits
      `);
      }
      return;
    }
    const { stderrCommit } = await gitCommit({ commit: commit2.trim() });
    if (stderrCommit)
      throw new CliError(`An error occured: ${stderrCommit}`);
    outro(dedent`
    You're all set 🎉

    use "git push" to publish your local commits
  `);
  } catch (err) {
    log({ type: "error", msg: err.message });
    handleCliError(err);
    process.exit(1);
  }
};

var update = command(
  {
    name: "update",
    description: "Update Commitizen CLI to the latest version"
  },
  async () => {
    const cmd = "npm update -g commitizen-cli";
    console.log(dim(`Running: ${cmd}`));
    const update = spawn(cmd, {
      stdio: "inherit",
      shell: process.env.SHELL || true
    });
    update.stderr?.on("error", () => {
      throw new CliError("An error occured on updating package");
    });
    update.on("close", (code) => {
      if (code !== 0) {
        console.log(`update process exited with code ${code}`);
      }
    });
  }
);

var config = command(
  {
    name: "config",
    parameters: ["[mode]", "[key=value...]"],
    description: "Get or Set Commitizen CLI config"
  },
  (argv) => {
    (async () => {
      const { mode, keyValue: keyValues } = argv._;
      if (!keyValues.length) {
        console.error('Error: Missing required parameter "key=value"\n');
        argv.showHelp();
        return process.exit(1);
      }
      if (mode === "get") {
        const config = await getConfig();
        for (const key of keyValues) {
          if (hasOwn(config, key)) {
            console.log(`${key}=${config[key]}`);
          }
        }
        return;
      }
      if (mode === "set") {
        await setConfigs(
          keyValues.map(
            (keyValue) => keyValue.split("=")
          )
        );
        return;
      }
      throw new CliError(`Invalid mode: ${mode}`);
    })().catch((err) => {
      log({ type: "error", msg: err.message });
      handleCliError(err);
      process.exit(1);
    });
  }
);

cli(
  {
    name: PACKAGE_NAME,
    version: VERSION,
    commands: [update, config]
  },
  () => {
    commiter().catch((err) => {
      log({ type: "error", msg: err.message });
      handleCliError(err);
      process.exit(1);
    });
  }
);
